/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::NJointController
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "NJointController.h"

#include <ArmarXCore/core/util/algorithm.h>

#include "../RobotUnit.h"

#include <atomic>

namespace armarx
{
    namespace RobotUnitModule
    {

        /**
         * \brief This class allows minimal access to private members of \ref armarx::RobotUnitModule::Devices in a sane fashion for \ref armarx::NJointController.
         * \warning !! DO NOT ADD ANYTHING IF YOU DO NOT KNOW WAHT YOU ARE DOING! IF YOU DO SOMETHING WRONG YOU WILL CAUSE UNDEFINED BEHAVIOUR !!
         */
        class DevicesAttorneyForNJointController
        {
            friend class ::armarx::NJointController;
            static JointController* GetJointController(Devices& d, const std::string& deviceName, const std::string& controlMode)
            {
                if (d.controlDevices.has(deviceName))
                {
                    auto& dev = d.controlDevices.at(deviceName);
                    if (dev->hasJointController(controlMode))
                    {
                        ARMARX_CHECK_NOT_NULL(dev->getJointController(controlMode));
                        ARMARX_CHECK_NOT_NULL(dev->getJointController(controlMode)->getControlTarget());
                        return dev->getJointController(controlMode);
                    }
                }
                return nullptr;
            }
        };
    }
}


thread_local bool armarx::NJointControllerRegistryEntry::ConstructorIsRunning_ = false;

namespace armarx
{
    const NJointControllerPtr NJointController::NullPtr
    {
        nullptr
    };

    NJointControllerDescription NJointController::getControllerDescription(const Ice::Current&) const
    {
        NJointControllerDescription d;
        d.className = getClassName();
        ARMARX_CHECK_EXPRESSION(getProxy(-1));
        d.controller = NJointControllerInterfacePrx::uncheckedCast(getProxy(-1));
        d.controlModeAssignment = controlDeviceControlModeMap;
        d.instanceName = getInstanceName();
        d.deletable = deletable;
        d.internal = internal;
        return d;
    }

    boost::optional<std::vector<char> > NJointController::isNotInConflictWith(const std::vector<char>& used) const
    {
        ARMARX_CHECK_EXPRESSION(used.size() == controlDeviceUsedBitmap.size());
        auto result = used;
        for (std::size_t i = 0; i < controlDeviceUsedBitmap.size(); ++i)
        {
            if (controlDeviceUsedBitmap.at(i))
            {
                if (used.at(i))
                {
                    return boost::optional<std::vector<char>>();
                }
                result.at(i) = true;
            }

        }
        return {true, std::move(result)};
    }

    NJointControllerStatus NJointController::getControllerStatus(const Ice::Current&) const
    {
        NJointControllerStatus s;
        s.active = isActive;
        s.instanceName = getInstanceName();
        s.requested = isRequested;
        s.error = deactivatedBecauseOfError;
        s.timestampUSec = std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::high_resolution_clock::now().time_since_epoch()).count();
        return s;
    }

    NJointControllerDescriptionWithStatus NJointController::getControllerDescriptionWithStatus(const Ice::Current&) const
    {
        NJointControllerDescriptionWithStatus ds;
        ds.description = getControllerDescription();
        ds.status = getControllerStatus();
        return ds;
    }

    RobotUnitInterfacePrx NJointController::getRobotUnit(const Ice::Current&) const
    {
        return RobotUnitInterfacePrx::uncheckedCast(robotUnit.getProxy(-1));
    }

    void NJointController::activateController(const Ice::Current&)
    {
        robotUnit.activateNJointController(this);
    }

    void NJointController::deactivateController(const Ice::Current&)
    {
        robotUnit.deactivateNJointController(this);
    }

    void NJointController::deleteController(const Ice::Current&)
    {
        robotUnit.deleteNJointController(this);
    }
    void NJointController::deactivateAndDeleteController(const Ice::Current&)
    {
        robotUnit.deactivateAndDeleteNJointController(this);
    }

    void NJointController::publish(const SensorAndControl& sac, const DebugDrawerInterfacePrx& draw, const DebugObserverInterfacePrx& observer)
    {
        const bool active = isActive;
        if (publishActive.exchange(active) != active)
        {
            if (active)
            {
                ARMARX_DEBUG << "activating publishing for " << getInstanceName();
                try
                {
                    onPublishActivation(draw, observer);
                }
                catch (std::exception& e)
                {
                    ARMARX_WARNING << "onPublishActivation of '" << getInstanceName() << "' threw an exception!\nWhat:\n" << e.what();
                }
                catch (...)
                {
                    ARMARX_WARNING << "onPublishActivation of '" << getInstanceName() << "' threw an exception!";
                }
            }
            else
            {
                ARMARX_DEBUG << "deactivating publishing for " << getInstanceName();
                try
                {
                    onPublishDeactivation(draw, observer);
                }
                catch (std::exception& e)
                {
                    ARMARX_WARNING << "onPublishDeactivation of '" << getInstanceName() << "' threw an exception!\nWhat:\n" << e.what();
                }
                catch (...)
                {
                    ARMARX_WARNING << "onPublishDeactivation of '" << getInstanceName() << "' threw an exception!";
                }
            }
        }
        if (active)
        {
            try
            {
                onPublish(sac, draw, observer);
            }
            catch (std::exception& e)
            {
                ARMARX_WARNING << "onPublish of '" << getInstanceName() << "' threw an exception!\nWhat:\n" << e.what();
            }
            catch (...)
            {
                ARMARX_WARNING << "onPublish of '" << getInstanceName() << "' threw an exception!";
            }
        }
    }

    void NJointController::deactivatePublish(const DebugDrawerInterfacePrx& draw, const DebugObserverInterfacePrx& observer)
    {
        if (publishActive.exchange(false))
        {
            ARMARX_DEBUG << "forced deactivation of publishing for " << getInstanceName();
            onPublishDeactivation(draw, observer);
        }
    }

    void NJointController::rtActivateController()
    {
        if (!isActive)
        {
            deactivatedBecauseOfError = false;
            rtPreActivateController();
            isActive = true;
            errorState.store(false);
        }
    }

    void NJointController::rtDeactivateController()
    {
        if (isActive)
        {
            isActive = false;
            rtPostDeactivateController();
        }
    }

    void NJointController::rtDeactivateControllerBecauseOfError()
    {
        deactivatedBecauseOfError = true;
        rtDeactivateController();
    }

    armarx::ConstSensorDevicePtr armarx::NJointController::peekSensorDevice(const std::string& deviceName) const
    {
        ARMARX_CHECK_EXPRESSION_W_HINT(
            NJointControllerRegistryEntry::ConstructorIsRunning(),
            "The function '" << BOOST_CURRENT_FUNCTION << "' must only be called during the construction of an NJointController."
        );
        return RobotUnitModule::Devices::Instance().getSensorDevice(deviceName);
    }

    ConstControlDevicePtr NJointController::peekControlDevice(const std::string& deviceName) const
    {
        ARMARX_CHECK_EXPRESSION_W_HINT(
            NJointControllerRegistryEntry::ConstructorIsRunning(),
            "The function '" << BOOST_CURRENT_FUNCTION << "' must only be called during the construction of an NJointController."
        );
        return RobotUnitModule::Devices::Instance().getControlDevice(deviceName);
    }

    void NJointController::useSynchronizedRtRobot(bool updateCollisionModel)
    {
        ARMARX_CHECK_EXPRESSION_W_HINT(
            NJointControllerRegistryEntry::ConstructorIsRunning(),
            "The function '" << BOOST_CURRENT_FUNCTION << "' must only be called during the construction of an NJointController."
        );
        ARMARX_CHECK_IS_NULL_W_HINT(rtRobot, "useSynchronizedRtRobot was already called");
        rtRobot = RobotUnitModule::RobotData::Instance().cloneRobot(updateCollisionModel);
        rtRobotNodes = rtRobot->getRobotNodes();
    }

    const SensorValueBase* NJointController::useSensorValue(const std::string& deviceName) const
    {
        ARMARX_CHECK_EXPRESSION_W_HINT(
            NJointControllerRegistryEntry::ConstructorIsRunning(),
            "The function '" << BOOST_CURRENT_FUNCTION << "' must only be called during the construction of an NJointController."
        );
        auto dev = peekSensorDevice(deviceName);
        if (!dev)
        {
            ARMARX_DEBUG << "No sensor device for " << deviceName;
            return nullptr;
        }
        return dev->getSensorValue();
    }

    NJointController::NJointController() :
        robotUnit(RobotUnitModule::ModuleBase::Instance< ::armarx::RobotUnit>()),
        controlDeviceUsedBitmap(robotUnit.getNumberOfControlDevices(), 0)
    {
        controlDeviceUsedIndices.reserve(robotUnit.getNumberOfControlDevices());
    }

    ControlTargetBase* NJointController::useControlTarget(const std::string& deviceName, const std::string& controlMode)
    {
        ARMARX_CHECK_EXPRESSION_W_HINT(
            NJointControllerRegistryEntry::ConstructorIsRunning(),
            "The function '" << BOOST_CURRENT_FUNCTION << "' must only be called during the construction of an NJointController."
        );
        ARMARX_CHECK_EQUAL_W_HINT(
            controlDeviceControlModeMap.count(deviceName), 0,
            BOOST_CURRENT_FUNCTION << ": Must not request two ControlTargets for the same device. (got = "
            << controlDeviceControlModeMap.at(deviceName) << ", requested " + controlMode + ") "
            << "(You can only have a ControlDevice in one ControlMode. "
            << "To check all available ControlModes for this device, get the ControlDevice via peekControlDevice(" + deviceName + ") and querry it)"
        );

        //there is a device and a target was requested:
        JointController* const jointCtrl = RobotUnitModule::DevicesAttorneyForNJointController::GetJointController(RobotUnitModule::Devices::Instance(), deviceName, controlMode);
        if (!jointCtrl)
        {
            return nullptr;
        }

        //update meta data structured
        const std::size_t devIndex = robotUnit.getControlDeviceIndex(deviceName);
        controlDeviceUsedJointController[deviceName] = jointCtrl;

        controlDeviceControlModeMap[deviceName] = controlMode;

        ARMARX_CHECK_EQUAL(0, controlDeviceUsedBitmap.at(devIndex));
        controlDeviceUsedBitmap.at(devIndex) = 1;
        //we want this vector to be sorted
        controlDeviceUsedIndices.insert
        (
            std::upper_bound(controlDeviceUsedIndices.begin(), controlDeviceUsedIndices.end(), devIndex),
            devIndex
        );

        return jointCtrl->getControlTarget();
    }
}
